#ifndef REF_COUNTED_HPP
#define REF_COUNTED_HPP

#include <iostream>
#include <type_traits>

template <typename T>
class ref_counted
{
private:
    T *object;      // Pointer to the managed object
    int *ref_count; // Pointer to the reference count

public:
    // Constructor: accepts dynamically allocated objects
    explicit ref_counted(T *obj) : object(obj), ref_count(new int(1)) {}

    // Copy constructor: increments reference count
    ref_counted(const ref_counted &other) : object(other.object), ref_count(other.ref_count)
    {
        (*ref_count)++;
    }

    // Assignment operator: manages reference count
    ref_counted &operator=(const ref_counted &other)
    {
        if (this != &other)
        {
            // Decrement current reference count
            if (--(*ref_count) == 0)
            {
                deleteObject();
                delete ref_count;
            }
            // Share new object's memory
            object = other.object;
            ref_count = other.ref_count;
            (*ref_count)++;
        }
        return *this;
    }

    // Destructor: decrements reference count and deletes object if necessary
    ~ref_counted()
    {
        if (--(*ref_count) == 0)
        {
            deleteObject();
            delete ref_count;
        }
    }

    // Access the managed object
    T *get() const { return object; }

    // Access the reference count
    int count() const { return *ref_count; }

private:
    // Helper function to delete object safely
    void deleteObject()
    {
        if constexpr (std::is_array_v<T>)
        {
            delete[] object; // Use delete[] for arrays
        }
        else
        {
            delete object; // Use delete for single objects
        }
    }
};

#endif
